-------------------------------------------------------------------------------
--! @file       JoltikBC.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Joltik BC
------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.JoltikBC_pkg.all;

entity JoltikBC is
    port    (
        ii        : in  std_logic_vector(64             -1 downto 0);
        rdkey     : in  std_logic_vector(64             -1 downto 0);
        oo        : out std_logic_vector(64             -1 downto 0)        
    );
end entity JoltikBC;

architecture EncRound of JoltikBC is
    type state is array (0 to 15) of std_logic_vector(3 downto 0);
    constant SBOX   : state := (x"E", x"4", x"B", x"2", x"3", x"8", x"0", x"9", x"1", x"A", x"7", x"F", x"6", x"C", x"5", x"D");
    signal ss       : state;
    signal sub      : state;
    signal shift    : state;
    signal mix      : state;
begin
    
    gMap:
    for i in 0 to 15 generate
        ss(i) <= ii(64-i*4-1 downto 64-(i+1)*4);
    end generate;
    
    gSub:
    for i in 0 to 15 generate
        sub(i) <= SBOX(to_integer(unsigned(ss(i))));
    end generate;
    
    gShift:
    for i in 0 to 3 generate
        gShift2:
        for j in 0 to 3 generate
            shift(i+j*4) <= sub(i+((j+i) mod 4)*4);
        end generate;
    end generate;
    
    gMix:
    for i in 0 to 3 generate
        mix(i*4+0) <=       shift(i*4)  xor mul4 (shift(i*4+1)) xor mul9 (shift(i*4+2)) xor mul13(shift(i*4+3));
        mix(i*4+1) <= mul4 (shift(i*4)) xor       shift(i*4+1)  xor mul13(shift(i*4+2)) xor mul9 (shift(i*4+3));
        mix(i*4+2) <= mul9 (shift(i*4)) xor mul13(shift(i*4+1)) xor       shift(i*4+2)  xor mul4 (shift(i*4+3));
        mix(i*4+3) <= mul13(shift(i*4)) xor mul9 (shift(i*4+1)) xor mul4 (shift(i*4+2)) xor       shift(i*4+3);
    end generate;
    
    gAdd:
    for i in 0 to 15 generate
        oo(64-i*4-1 downto 64-(i+1)*4) <= mix(i) xor rdkey(64-i*4-1 downto 64-(i+1)*4);
    end generate;    
end architecture EncRound;

architecture DecRound of JoltikBC is
    type state is array (0 to 15) of std_logic_vector(3 downto 0);
    constant INVSBOX   : state := (x"6", x"8", x"3", x"4", x"1", x"E", x"C", x"A", x"5", x"7", x"9", x"2", x"D", x"F", x"0", x"B");
    signal ss       : state;
    signal mix      : state;
    signal shift    : state;
    signal sub      : state;
begin
    
    gMap:
    for i in 0 to 15 generate
        ss(i) <= ii(64-i*4-1 downto 64-(i+1)*4);
    end generate;
    
    gInvMix:
    for i in 0 to 3 generate
        mix(i*4+0) <=       ss(i*4)  xor mul4 (ss(i*4+1)) xor mul9 (ss(i*4+2)) xor mul13(ss(i*4+3));
        mix(i*4+1) <= mul4 (ss(i*4)) xor       ss(i*4+1)  xor mul13(ss(i*4+2)) xor mul9 (ss(i*4+3));
        mix(i*4+2) <= mul9 (ss(i*4)) xor mul13(ss(i*4+1)) xor       ss(i*4+2)  xor mul4 (ss(i*4+3));
        mix(i*4+3) <= mul13(ss(i*4)) xor mul9 (ss(i*4+1)) xor mul4 (ss(i*4+2)) xor       ss(i*4+3);
    end generate;
    
    gInvShift:
    for i in 0 to 3 generate
        gShift2:
        for j in 0 to 3 generate
            shift(i+j*4) <= mix(i+((j+4-i) mod 4)*4);
        end generate;
    end generate;
        
    gInvSub:
    for i in 0 to 15 generate
        sub(i) <= INVSBOX(to_integer(unsigned(shift(i))));
    end generate;    
    
    gAdd:
    for i in 0 to 15 generate
        oo(64-i*4-1 downto 64-(i+1)*4) <= sub(i) xor rdkey(64-i*4-1 downto 64-(i+1)*4);
    end generate;    
end architecture DecRound;