-------------------------------------------------------------------------------
--! @file       JoltikBC_pkg.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Joltik package
------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

package JoltikBC_pkg is
    constant MSB_AD                 : std_logic_vector(3 downto 0) := "0010";
    constant MSB_AD_LAST            : std_logic_vector(3 downto 0) := "0110";
    constant MSB_M                  : std_logic_vector(3 downto 0) := "0000";
    constant MSB_M_LAST_NONZERO     : std_logic_vector(3 downto 0) := "0100";
    constant MSB_CHKSUM_FULL        : std_logic_vector(3 downto 0) := "0001";
    constant MSB_CHKSUM_NON_FULL    : std_logic_vector(3 downto 0) := "0101";
    
    function  mul2(x: std_logic_vector) return std_logic_vector;
    function  mul4(x: std_logic_vector) return std_logic_vector;
    function  mul8(x: std_logic_vector) return std_logic_vector;
    function  mul9(x: std_logic_vector) return std_logic_vector;
    function mul13(x: std_logic_vector) return std_logic_vector;
end package JoltikBC_pkg;

package body JoltikBC_pkg is
    function mul2(x: std_logic_vector) return std_logic_vector is
        variable y : std_logic_vector(3 downto 0);
    begin
        y(3) :=                   x(2);
        y(2) :=          x(1);
        y(1) := x(0)                   xor x(3);
        y(0) :=                            x(3);
        return y;
    end function mul2;
    
    function mul4(x: std_logic_vector) return std_logic_vector is
        variable y : std_logic_vector(3 downto 0);
    begin
        y(3) :=          x(1);
        y(2) := x(0)                   xor x(3);
        y(1) :=                   x(2) xor x(3);
        y(0) :=                   x(2);
        return y;
    end function mul4;
    
    function mul8(x: std_logic_vector) return std_logic_vector is
        variable y : std_logic_vector(3 downto 0);
    begin
        y(3) := x(0)                   xor x(3);
        y(2) :=                   x(2) xor x(3);
        y(1) :=          x(1) xor x(2);
        y(0) :=          x(1);
        return y;
    end function mul8;

    function mul9(x: std_logic_vector) return std_logic_vector is
        variable y : std_logic_vector(3 downto 0);
    begin
        y(3) := x(0);
        y(2) :=                            x(3);
        y(1) :=                    x(2);
        y(0) := x(0) xor x(1);
        return y;
    end function mul9;

    function mul13(x: std_logic_vector) return std_logic_vector is
        variable y : std_logic_vector(3 downto 0);
    begin
        y(3) := x(0) xor x(1);
        y(2) := x(0);
        y(1) :=                            x(3);
        y(0) := x(0) xor x(1) xor x(2);
        return y;    
    end function mul13;    
end package body JoltikBC_pkg;